@@ -1,5 +1,5 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+   Copyright (C) 2000,2001,2002,2003,2004 Free Software Foundation, Inc.
 
    Written by Sergey Poznyakoff
   
@@ -17,20 +17,47 @@
    along with GNU Radius; if not, write to the Free Software Foundation, 
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
-#include <sysdep.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <stdio.h>
-#include <radius.h>
-#include <radpaths.h>
 #include <signal.h>
+
+#include <common.h>
 #include <cfg.h>
-#include <list.h>
+#include <radsql.h>
 
 /* Server data structures */
 struct radutmp; /* declared in radutmp.h */
-struct obstack;
+
+/* Struct radiusd_request expands grad_request_t by adding server-specific
+   fields to it. */
+typedef struct radiusd_request {
+	grad_request_t *request;
+	
+        /* Saved reply values */
+        int           reply_code;   /* Reply code */
+        grad_avp_t    *reply_pairs; /* Reply pairs */
+        char          *reply_msg;   /* Reply message */
+	                            /* FIXME: should probably be
+			 	       incorporated to reply_pairs
+				       at once */
+	/* List of cfg file locations that lead to the decision on this
+	   request */
+	grad_list_t   *locus_list;
+	
+        /* Proxy support fields */
+        grad_realm_t  *realm;       
+        int           validated;     /* Already md5 checked */
+	int           server_no;
+	int           attempt_no;
+        grad_uint32_t server_id;     /* Proxy ID of the packet */
+	char          *remote_user;  /* Remote username */
+        u_char        remote_auth[GRAD_AUTHENTICATOR_LENGTH];
+	                             /* Remote request authenticator */	
+        int           server_code;   /* Reply code from other srv */
+        grad_avp_t    *server_reply; /* Reply from other server */
+} radiusd_request_t;
 
 enum reload_what {
         reload_config,
@@ -115,18 +142,18 @@ struct queue_stat {
 typedef struct queue_stat QUEUE_STAT[R_MAX];
 
 typedef struct client {
-	NETDEF                  netdef;
-        char                    longname[MAX_LONGNAME+1];
+	grad_netdef_t           netdef;
+        char                    longname[GRAD_MAX_LONGNAME+1];
         u_char                  *secret;
-        char                    shortname[MAX_SHORTNAME+1];
+        char                    shortname[GRAD_MAX_SHORTNAME+1];
 } CLIENT;
 
 typedef struct proxy_state {
-	UINT4 ref_ip;             /* Radius server IP */
-        UINT4 proxy_id;           /* Proxy ID assigned by the server */
-        UINT4 remote_ip;          /* Remote radius server IP */
-        UINT4 client_ip;          /* IP of the requesting client */ 
-        UINT4 id;                 /* Radius request ID */
+	grad_uint32_t ref_ip;             /* Radius server IP */
+        grad_uint32_t proxy_id;           /* Proxy ID assigned by the server */
+        grad_uint32_t remote_ip;          /* Remote radius server IP */
+        grad_uint32_t client_ip;          /* IP of the requesting client */ 
+        grad_uint32_t id;                 /* Radius request ID */
 } PROXY_STATE;
 
 typedef struct {
@@ -142,9 +169,9 @@ typedef struct {
 typedef struct user_symbol {
         struct user_symbol *next;  /* Link to the next entry */
         char *name;                /* Label */
-        VALUE_PAIR *check;         /* LHS */
-        VALUE_PAIR *reply;         /* RHS */
-        LOCUS loc;                 /* Location of the definition of
+        grad_avp_t *check;         /* LHS */
+        grad_avp_t *reply;         /* RHS */
+        grad_locus_t loc;                 /* Location of the definition of
 				      this entry */
         int ordnum;                /* Entry ordinal number (used for semantic
 				      checks */
@@ -160,7 +187,7 @@ typedef struct user_symbol {
 typedef struct netname NETNAME;
 struct netname {
 	char *name;
-	RAD_LIST /* of NETDEF */ *netlist;
+	grad_list_t /* of grad_netdef_t */ *netlist;
 };
 
 typedef struct community Community;
@@ -172,13 +199,13 @@ struct community {
 typedef struct access_control_list ACL;
 struct access_control_list {
         Community *community;   /* community or NULL to deny access */
-	RAD_LIST /* of NETDEF */ *netlist;
+	grad_list_t /* of grad_netdef_t */ *netlist;
 };
 
 struct radstat {
         struct timeval start_time;
-        counter port_active_count;
-        counter port_idle_count;
+        grad_counter_t port_active_count;
+        grad_counter_t port_idle_count;
 };
 
 typedef enum {
@@ -199,9 +226,9 @@ typedef struct {
 
 #define stat_inc(m,a,c) \
  do if (server_stat) {\
-        NAS *nas;\
+        grad_nas_t *nas;\
         server_stat -> m . c ++;\
-        if ((nas = nas_lookup_ip(a)) != NULL && nas->app_data)\
+        if ((nas = grad_nas_lookup_ip(a)) != NULL && nas->app_data)\
                 ((struct nas_stat*)nas->app_data)-> m . c ++;\
  } while (0)
 
@@ -226,8 +253,8 @@ typedef void (*config_hook_fp)(void *func_data, void *app_data);
 #define MAX_REQUESTS            255
 #define MAX_CHILDREN            16
 #define PROCESS_TIMEOUT         3600
-#define RADIUSD_READ_TIMEOUT    0
-#define RADIUSD_WRITE_TIMEOUT   0
+#define RADIUSD_READ_TIMEOUT    5
+#define RADIUSD_WRITE_TIMEOUT   5
 
 /*
  * Authentication results
@@ -280,6 +307,9 @@ extern int radius_mode;
 extern int debug_flag;
 extern int auth_detail;
 extern int acct_detail;
+extern char *auth_detail_template;
+extern char *acct_detail_template;
+extern int acct_system;
 extern int auth_trace_rules;
 extern int acct_trace_rules;
 extern int strip_names;
@@ -290,11 +320,11 @@ extern size_t max_children;
 extern unsigned process_timeout;
 extern unsigned radiusd_write_timeout;
 extern unsigned radiusd_read_timeout;
-extern UINT4 expiration_seconds;
-extern UINT4 warning_seconds;
+extern grad_uint32_t expiration_seconds;
+extern grad_uint32_t warning_seconds;
 extern int use_dbm;
-extern UINT4 myip;
-extern UINT4 ref_ip;
+extern grad_uint32_t myip;
+extern grad_uint32_t ref_ip;
 extern int auth_port;
 extern int acct_port;
 extern int suspend_flag;
@@ -321,6 +351,7 @@ extern int auth_comp_flag;
 extern int acct_comp_flag; 
 extern RADIUS_USER exec_user;
 extern RADIUS_USER radiusd_user;
+extern grad_symtab_t *user_tab;
 
 /* Input subsystem (input.c) */
 
@@ -346,8 +377,10 @@ void input_iterate_channels(INPUT *input, char *name, list_iterator_t fun,
 /* rpp.c */
 int rpp_ready();
 int rpp_forward_request(REQUEST *req);
-void rpp_remove(pid_t pid);
+void rpp_remove_pid(pid_t pid);
+void rpp_status_changed(pid_t pid, int exit_status);
 void rpp_flush(int (*fun)(void*), void *closure);
+void rpp_collect_exited();
 int rpp_input_handler(int fd, void *data);
 int rpp_input_close(int fd, void *data);
 int rpp_kill(pid_t pid, int signo);
@@ -372,7 +405,7 @@ int udp_input_handler(int fd, void *data);
 int udp_input_close(int fd, void *data);
 int udp_input_cmp(const void *a, const void *b);
 
-int udp_open(int type, UINT4 ipaddr, int port, int nonblock);
+int udp_open(int type, grad_uint32_t ipaddr, int port, int nonblock);
 
 void radiusd_pidfile_write(char *name);
 pid_t radiusd_pidfile_read(char *name);
@@ -395,22 +428,21 @@ void radiusd_close_channel(int fd);
 /* exec.c */
 int radius_get_user_ids(RADIUS_USER *usr, const char *name);
 int radius_switch_to_user(RADIUS_USER *usr);
-int radius_exec_program(char *, RADIUS_REQ *, VALUE_PAIR **, int);
+int radius_exec_command(char *cmd);
+int radius_exec_program(char *, radiusd_request_t *, grad_avp_t **, int);
 void filter_cleanup(pid_t pid, int status);
-int filter_auth(char *name, RADIUS_REQ *req, VALUE_PAIR **reply_pairs);
-int filter_acct(char *name, RADIUS_REQ *req);
+int filter_auth(char *name, radiusd_request_t *req, grad_avp_t **reply_pairs);
+int filter_acct(char *name, radiusd_request_t *req);
 int filters_stmt_term(int finish, void *block_data, void *handler_data);
 extern struct cfg_stmt filters_stmt[];
 
 /* scheme.c */
 void scheme_main();
-void scheme_load(char *filename);
 void scheme_load_path(char *pathname);
-void scheme_debug(int val);
-int scheme_auth(char *procname, RADIUS_REQ *req,
-                VALUE_PAIR *user_check, VALUE_PAIR **user_reply_ptr);
-int scheme_acct(char *procname, RADIUS_REQ *req);
-void scheme_add_load_path(char *path);
+int scheme_auth(char *procname, radiusd_request_t *req,
+                grad_avp_t *user_check, grad_avp_t **user_reply_ptr);
+int scheme_acct(char *procname, radiusd_request_t *req);
+int scheme_eval_boolean_expr(char *expr);
 void scheme_read_eval_loop();
 void scheme_redirect_output();
 void start_guile();
@@ -419,7 +451,13 @@ int guile_cfg_handler(int argc, cfg_value_t *argv,
 extern struct cfg_stmt guile_stmt[];
 
 /* log.c */
-void sqllog __PVAR((int status, char *msg, ...));
+void radiusd_logger(int level,
+		    const grad_request_t *req,
+		    const grad_locus_t *loc,
+		    const char *func_name,
+		    int en,
+		    const char *fmt, va_list ap);
+void sqllog(int status, char *msg, ...);
 int logging_stmt_handler(int argc, cfg_value_t *argv, void *block_data,
 			 void *handler_data);
 int logging_stmt_end(void *block_data, void *handler_data);
@@ -434,8 +472,8 @@ extern struct cfg_stmt logging_stmt[];
 #define REQ_AUTH_ZERO 1
 #define REQ_AUTH_BAD  2
 
-void radius_send_reply(int, RADIUS_REQ *, VALUE_PAIR *, char *, int);
-void radius_send_challenge(RADIUS_REQ *radreq, char *msg, char *state, int fd);
+void radius_send_reply(int, radiusd_request_t *, grad_avp_t *, char *, int);
+void radius_send_challenge(radiusd_request_t *radreq, char *msg, char *state, int fd);
 int radius_verify_digest(REQUEST *req);
 
 int radius_auth_req_decode(struct sockaddr_in *sa,
@@ -450,10 +488,10 @@ void radius_req_xmit(REQUEST *request);
 int radius_req_failure(int type, struct sockaddr_in *addr);
 void radius_req_update(void *req_ptr, void *data_ptr);
 int radius_respond(REQUEST *req);
-void radius_req_register_locus(RADIUS_REQ *req, LOCUS *loc);
-void radius_trace_path(RADIUS_REQ *req);
-VALUE_PAIR *radius_decrypt_request_pairs(RADIUS_REQ *req, VALUE_PAIR *pair);
-void radius_destroy_pairs(VALUE_PAIR **p);
+void radius_req_register_locus(radiusd_request_t *req, grad_locus_t *loc);
+void radius_trace_path(radiusd_request_t *req);
+grad_avp_t *radius_decrypt_request_pairs(radiusd_request_t *req, grad_avp_t *pair);
+void radius_destroy_pairs(grad_avp_t **p);
 
 /* shmem.c */
 int shmem_alloc(size_t size);
@@ -466,11 +504,11 @@ int pam_pass(char *name, char *passwd, const char *pamauth, char **reply_msg);
 
 /* proxy.c */
 int proxy_send(REQUEST *req);
-int proxy_receive(RADIUS_REQ *radreq, RADIUS_REQ *oldreq, int activefd);
-void proxy_retry(RADIUS_REQ *radreq, int fd);
-int proxy_cmp(RADIUS_REQ *qr, RADIUS_REQ *r);
-VALUE_PAIR *proxy_request_recode(RADIUS_REQ *radreq, VALUE_PAIR *plist,
-				 u_char *secret, u_char *vector);
+int proxy_receive(radiusd_request_t *radreq, radiusd_request_t *oldreq, int activefd);
+void proxy_retry(radiusd_request_t *radreq, int fd);
+int proxy_cmp(radiusd_request_t *qr, radiusd_request_t *r);
+grad_avp_t *proxy_request_recode(radiusd_request_t *radreq, grad_avp_t *plist,
+				 u_char *secret, u_char *authenticator);
 
 /* menu.c */
 #define MAX_PATH_LENGTH                 256
@@ -480,50 +518,74 @@ VALUE_PAIR *proxy_request_recode(RADIUS_REQ *radreq, VALUE_PAIR *plist,
 #define MAX_STATE_VALUE                 128
 #define RAD_BUFFER_SIZE                 4096
 
-void menu_reply(RADIUS_REQ *radreq, int fd);
+void menu_reply(radiusd_request_t *radreq, int fd);
 char *menu_read_text(char *menu_name);
 
 /* acct.c */
+void system_acct_init();
 void acct_init();
-int rad_accounting(RADIUS_REQ *, int, int);
-int radzap(UINT4 nas, int port, char *user, time_t t);
-int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul,
-		    int *pcount);
-int rad_check_realm(REALM *realm);
-int write_detail(RADIUS_REQ *radreq, int authtype, char *f);
+int rad_accounting(radiusd_request_t *, int, int);
+int radzap(grad_uint32_t nas, int port, char *user, time_t t);
+int write_detail(radiusd_request_t *radreq, int authtype, int rtype);
+
+int radutmp_mlc_collect_user(char *name, radiusd_request_t *request,
+			     grad_list_t **sess_list);
+int radutmp_mlc_collect_realm(radiusd_request_t *request,
+			      grad_list_t **sess_list);
+void radutmp_mlc_close(struct radutmp *up);
+int radutmp_mlc_enabled_p(void);
+
+/* mlc.c */
+typedef int (*mlc_collect_user_t) (char *name, radiusd_request_t *request,
+				   grad_list_t **sess_list);
+typedef int (*mlc_collect_realm_t) (radiusd_request_t *request,
+				    grad_list_t **sess_list);
+typedef void (*mlc_close_t) (struct radutmp *up);
+typedef int (*mlc_enabled_t) (void);
+void mlc_register_method(char *name,
+			 mlc_collect_user_t collect_user,
+			 mlc_collect_realm_t collect_realm,
+			 mlc_close_t close,
+			 mlc_enabled_t enabled_p);
+
+extern struct cfg_stmt mlc_stmt[];
+int radius_mlc_user(char *name, radiusd_request_t *request,
+		    size_t maxsimul, size_t *pcount);
+int radius_mlc_realm(radiusd_request_t *request);
 
 /* files.c */
-int user_find(char *name, RADIUS_REQ *, VALUE_PAIR **, VALUE_PAIR **);
-int userparse(char *buf, VALUE_PAIR **first_pair, char **errmsg);
-void presuf_setup(VALUE_PAIR *request_pairs);
-int hints_setup(RADIUS_REQ *request);
-int huntgroup_access(RADIUS_REQ *radreq, LOCUS *loc);
-CLIENT *client_lookup_ip(UINT4 ipno);
-char *client_lookup_name(UINT4 ipno, char *buf, size_t size);
+int user_find(char *name, radiusd_request_t *, grad_avp_t **, grad_avp_t **);
+int userparse(char *buf, grad_avp_t **first_pair, char **errmsg);
+void presuf_setup(grad_avp_t *request_pairs);
+int hints_setup(radiusd_request_t *request);
+int huntgroup_access(radiusd_request_t *radreq, grad_locus_t *loc);
+CLIENT *client_lookup_ip(grad_uint32_t ipno);
+char *client_lookup_name(grad_uint32_t ipno, char *buf, size_t size);
 int read_clients_file(char *);
-NAS *nas_find(UINT4 ipno);
-NAS *nas_by_name(char *name);
-char *nas_name(UINT4 ipno);
-char *nas_name2(RADIUS_REQ *r);
-int read_naslist_file(char *);
+grad_nas_t *grad_nas_find(grad_uint32_t ipno);
+grad_nas_t *grad_nas_by_name(char *name);
+char *grad_nas_name(grad_uint32_t ipno);
+char *nas_name2(radiusd_request_t *r);
+int read_nas_list_file(char *);
 int reload_config_file(enum reload_what);
-int presufcmp(VALUE_PAIR *check, char *name, char *rest);
+int presufcmp(grad_avp_t *check, char *name, char *rest);
 int get_config();
 int get_deny(char *user);
-NAS *findnasbyindex(int);
+grad_nas_t *findnasbyindex(int);
 char *make_server_ident();
 void dump_users_db();
 void strip_username(int do_strip, char *name,
-                    VALUE_PAIR *check_item, char *stripped_name);
+                    grad_avp_t *check_item, char *stripped_name);
 
 /* version.c */
 void version();
 
 /* auth.c */
-int rad_auth_init(RADIUS_REQ *radreq, int activefd);
-int rad_auth_check_username(RADIUS_REQ *radreq, int activefd);
-int rad_authenticate (RADIUS_REQ *, int);
-void req_decrypt_password(char *password, RADIUS_REQ *req, VALUE_PAIR *pair);
+int rad_auth_init(radiusd_request_t *radreq, int activefd);
+int rad_auth_check_username(radiusd_request_t *radreq, int activefd);
+int rad_authenticate (radiusd_request_t *, int);
+void req_decrypt_password(char *password, grad_request_t *req,
+			  grad_avp_t *pair);
 
 /* timestr.c */
 int timestr_match(char *, time_t);
@@ -533,7 +595,7 @@ int timestr_match(char *, time_t);
 void snmpserv_init(void *arg);
 void snmp_auth_server_reset();
 void snmp_acct_server_reset();
-void snmp_attach_nas_stat(NAS *nas);
+void snmp_attach_nas_stat(grad_nas_t *nas);
 void snmp_init_nas_stat();
 void snmp_sort_nas_stat();
 int snmp_stmt_begin(int finish, void *data, void *up_data);
@@ -546,9 +608,9 @@ void stat_init();
 void stat_done();
 void stat_update(struct radutmp *ut, int status);
 void stat_count_ports();
-struct nas_stat * find_nas_stat(UINT4 ip_addr);
-int stat_get_port_index(NAS *nas, int port_no);
-int stat_get_next_port_no(NAS *nas, int port_no);
+struct nas_stat * find_nas_stat(grad_uint32_t ip_addr);
+int stat_get_port_index(grad_nas_t *nas, int port_no);
+int stat_get_next_port_no(grad_nas_t *nas, int port_no);
 #else
 # define stat_init()
 # define stat_done()
@@ -565,21 +627,25 @@ void snmp_req_drop(int type, void *data, void *orig_data,
 int snmp_req_respond(REQUEST *request);
         
 /* radutil.c */
+radiusd_request_t *radiusd_request_alloc(grad_request_t *req);
+void radiusd_request_free(radiusd_request_t *radreq);
+
 char *radius_xlate(struct obstack *obp, char *str,
-                   RADIUS_REQ *req, VALUE_PAIR *reply_pairs);
-int radius_eval_avl(RADIUS_REQ *req, VALUE_PAIR *p);
+                   grad_request_t *req, grad_avp_t *reply_pairs);
+int radius_eval_avl(radiusd_request_t *req, grad_avp_t *p);
+char *util_xlate(struct obstack *sp, char *fmt, grad_request_t *radreq);
 
 /* rewrite.y */
 extern struct cfg_stmt rewrite_stmt[];
 int parse_rewrite(char *name);
 
 /* radck.c */
-int fix_check_pairs(int sf_file, LOCUS *loc, char *name, VALUE_PAIR **pairs);
-int fix_reply_pairs(int cf_file, LOCUS *loc, char *name, VALUE_PAIR **pairs);
+int fix_check_pairs(int sf_file, grad_locus_t *loc, char *name, grad_avp_t **pairs);
+int fix_reply_pairs(int cf_file, grad_locus_t *loc, char *name, grad_avp_t **pairs);
 void radck();
 
 /* checkrad.c */
-int checkrad(NAS *nas, struct radutmp *up);
+int checkrad(grad_nas_t *nas, struct radutmp *up);
 
 /* forward.c */
 int rad_cfg_forward_auth(int argc, cfg_value_t *argv,
@@ -587,6 +653,90 @@ int rad_cfg_forward_auth(int argc, cfg_value_t *argv,
 int rad_cfg_forward_acct(int argc, cfg_value_t *argv,
 			 void *block_data, void *handler_data);
 void forward_init();
-void forward_request(int type, RADIUS_REQ *req);
+void forward_request(int type, radiusd_request_t *req);
 
+/* dynload.c */
+extern struct cfg_stmt dynload_stmt[];
+int dynload_stmt_term(int finish, void *block_data, void *handler_data);
+void *radiusd_load_ext(const char *name, const char *ident, void **symbol);
+void dynload_init();
 
+/* Logging */
+
+/* log output modes */
+#define LM_UNKNOWN -1
+#define LM_OFF 0
+#define LM_FILE 1
+#define LM_SYSLOG 2
+
+/* log options */
+#define LO_CONS  0x0001
+#define LO_PID   0x0002
+#define LO_CAT   0x0004
+#define LO_PRI   0x0008
+#define LO_MSEC  0x0010
+#define LO_PERSIST 0x8000
+
+#define RADIUS_DEBUG_BUFFER_SIZE 1024
+
+typedef struct channel Channel;
+
+struct channel {
+        char *name;
+        int  pmask[L_NCAT]; 
+        int mode;   /* LM_ constant */
+        union {
+                int prio;        /* syslog: facility|priority */
+                char *file;      /* file: output file name */
+        } id;
+        int options;
+	char *prefix_hook;       /* prefix hook function */
+	char *suffix_hook;       /* suffix hook function */
+};
+
+Channel *channel_lookup(char *name);
+void channel_free(Channel *chan);
+void channel_free_list(Channel *chan);
+Channel * log_mark();
+void log_release();
+
+void register_channel(Channel *chan);
+void register_category(int cat, int pri, grad_list_t *chanlist);
+
+void log_set_to_console();
+void log_set_default(char *name, int cat, int pri);
+
+void log_open(int cat);
+void log_close();
+
+/* sql.c */
+#ifdef USE_SQL
+void sql_init();
+int radiusd_sql_config();
+void radiusd_sql_shutdown();
+void radiusd_sql_clear_cache();
+void radiusd_sql_acct(radiusd_request_t *req);
+int radiusd_sql_checkgroup(radiusd_request_t *req, char *groupname);
+int radiusd_sql_check_attr_query(radiusd_request_t *req, grad_avp_t **check_pairs);
+int radiusd_sql_reply_attr_query(radiusd_request_t *req, grad_avp_t **reply_pairs);
+void radiusd_sql_auth_result_query(radiusd_request_t *req, int fail);
+void radiusd_sql_cleanup(int type, void *req);
+
+char *radiusd_sql_pass(radiusd_request_t *req, char *data);
+
+# ifdef RADIUS_SERVER_GUILE
+SCM sql_exec_query(int type, char *query);
+# endif
+
+#else
+# define sql_init()
+# define radiusd_sql_config() 0
+# define radiusd_sql_shutdown()
+# define radiusd_sql_clear_cache()
+# define radiusd_sql_acct(req)
+# define radiusd_sql_checkgroup(req, groupname) 0
+# define radiusd_sql_check_attr_query(req, check_pairs) 0
+# define radiusd_sql_reply_attr_query(req, reply_pairs)
+# define radiusd_sql_auth_result_query(req, fail)
+# define radiusd_sql_cleanup (void (*)(int, void *)) NULL
+#endif
