@@ -1,5 +1,5 @@
 /* This file is part of GNU Radius
-   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+   Copyright (C) 2000,2001,2002,2003,2004 Free Software Foundation, Inc.
 
    Written by Sergey Poznyakoff
  
@@ -17,8 +17,6 @@
    along with GNU Radius; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
-#define RADIUS_MODULE_AUTH_C
-
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
@@ -66,12 +64,7 @@
 #endif
 
 #include <radiusd.h>
-#if defined(USE_SQL)
-# include <radsql.h>
-#endif
 #include <timestr.h>
-#include <envar.h>
-#include <obstack1.h>
 #include <rewrite.h>
 
 char *username_valid_chars;
@@ -89,6 +82,43 @@ check_user_name(char *p)
 
 LOCK_DECLARE(lock)
 
+#ifdef EMBEDDED_EXPIRATION_INFO	
+static char base64[] =         /* 0 ... 63 => ascii - 64 */
+        "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
+
+static int
+from_base64(u_char *data, int len, unsigned int *value)
+{
+	int i;
+
+	*value = 0;
+	for (i = len-1; i >= 0; i--) {
+		char *p = strchr(base64, data[i]);
+		if (!p)
+			return 1;
+		*value <<= 6;
+		*value += p - base64;
+	}
+	return 0;
+}
+	
+static int
+decode_aging_info(char *p, u_int *maxweeks, u_int *minweeks, u_int *lastchange)
+{
+	if (from_base64(p, 1, maxweeks))
+		return 1;
+	p++;
+	if (from_base64(p, 1, minweeks))
+		return 1;
+	p++;
+	if (from_base64(p, 2, lastchange))
+		return 1;
+	if (p[2])
+		return 1;
+	return 0;
+}
+#endif	
+	
 static enum auth_status
 unix_expiration(char *name, time_t *exp)
 {
@@ -133,6 +163,45 @@ unix_expiration(char *name, time_t *exp)
 		*exp = t;
 # endif
 	}
+#endif
+#ifdef EMBEDDED_EXPIRATION_INFO
+	if (status == auth_ok) {
+		struct passwd *pwd;
+		char *p;
+#define SECS_IN_WEEK 604800
+	
+		if (pwd = getpwnam(name)) {
+			p = strchr(pwd->pw_passwd, ',');
+			if (p) {
+				u_int maxweeks, minweeks, lastchange;
+			
+				if (decode_aging_info(p+1,
+						      &maxweeks,
+						      &minweeks,
+						      &lastchange) == 0) {
+					time_t now = time(NULL);
+					time_t nweeks = now / SECS_IN_WEEK
+						        - lastchange;
+					
+					if (maxweeks == minweeks)
+						return auth_password_expired;
+				
+					if (nweeks >= minweeks
+					    && nweeks >= maxweeks)
+						return auth_password_expired;
+
+					*exp = (maxweeks - nweeks)
+						* SECS_IN_WEEK
+						+ now % SECS_IN_WEEK;
+					status = auth_valid;
+				} else { /* Invalid password data? */
+					grad_log(L_NOTICE,
+						 _("Invalid password aging information for user '%s'"), name);
+					status = auth_fail;
+				}
+			}
+		}
+	}
 #endif
 	return status;
 }
@@ -160,9 +229,9 @@ unix_pass(char *name, char *passwd)
 	if (encrypted_pass) {
 		/* Check if the account is locked. */
 		if (SHADOW_PASSWD_LOCK(spwd) != 1) {
-			radlog(L_NOTICE,
-			       "unix_pass: [%s]: %s",
-			       name, _("account locked"));
+			grad_log(L_NOTICE,
+			         "unix_pass: [%s]: %s",
+			         name, _("account locked"));
 			encrypted_pass = NULL;
 		}
 	}
@@ -172,7 +241,7 @@ unix_pass(char *name, char *passwd)
 		if (encrypted_pass[0] == 0)
 			encrypted_pass = NULL;
 		else
-			encrypted_pass = estrdup(encrypted_pass);
+			encrypted_pass = grad_estrdup(encrypted_pass);
 	}
 	
 	LOCK_RELEASE(lock);
@@ -183,276 +252,73 @@ unix_pass(char *name, char *passwd)
         /*
          * Check encrypted password.
          */
-        pwlen = strlen(encrypted_pass)+1;
-        encpw = emalloc(pwlen);
-        rc = md5crypt(passwd, encrypted_pass, encpw, pwlen) == NULL
-                || strcmp(encpw, encrypted_pass);
-        efree(encpw);
-	efree(encrypted_pass);
+        pwlen = strlen(encrypted_pass);
+        encpw = grad_emalloc(pwlen+1);
+#ifdef EMBEDDED_EXPIRATION_INFO
+ {
+	/* Some systems append expiration data to the encrypted password */
+	char *p = strchr(encrypted_pass, ',');
+	if (p) 
+		pwlen = p - encrypted_pass;
+ }
+#endif
+        rc = grad_md5crypt(passwd, encrypted_pass, encpw, pwlen+1) == NULL
+                || strlen (encpw) != pwlen
+		|| memcmp(encpw, encrypted_pass, pwlen);
+        grad_free(encpw);
+	grad_free(encrypted_pass);
         if (rc)
                 return -1;
 
         return 0;
 }
 
-
-/* Check password. */
-static enum auth_status 
-rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
-                   VALUE_PAIR *namepair, char **user_msg,
-		   char *userpass,
-		   time_t *exp)
-{
-        char *ptr;
-        char *real_password = NULL;
-        char name[AUTH_STRING_LEN];
-        VALUE_PAIR *auth_item;
-        VALUE_PAIR *tmp;
-        int auth_type = -1;
-        int length;
-        enum auth_status result = auth_ok;
-        char *authdata = NULL;
-        char pw_digest[AUTH_VECTOR_LEN];
-        int pwlen;
-        char *pwbuf;
-        char *challenge;
-	int challenge_len;
-	
-        userpass[0] = 0;
-
-        /* Process immediate authentication types */
-        if ((tmp = avl_find(check_item, DA_AUTH_TYPE)) != NULL)
-                auth_type = tmp->avp_lvalue;
-
-	switch (auth_type) {
-	case DV_AUTH_TYPE_ACCEPT:
-                return auth_ok;
-
-	case DV_AUTH_TYPE_REJECT:
-                return auth_reject;
-
-	case DV_AUTH_TYPE_IGNORE:
-		return auth_ignore;
-	}
-
-        /* Find the password sent by the user. If it's not present,
-           authentication fails. */
-        
-        if (auth_item = avl_find(radreq->request, DA_CHAP_PASSWORD))
-                auth_type = DV_AUTH_TYPE_LOCAL;
-        else
-                auth_item = avl_find(radreq->request, DA_USER_PASSWORD);
-        
-        /* Decrypt the password. */
-        if (auth_item) {
-                if (auth_item->avp_strlength == 0)
-                        userpass[0] = 0;
-                else
-                        req_decrypt_password(userpass, radreq,
-                                             auth_item);
-        } else /* if (auth_item == NULL) */
-                return auth_fail;
-        
-        /* Set up authentication data */
-        if ((tmp = avl_find(check_item, DA_AUTH_DATA)) != NULL) 
-                authdata = tmp->avp_strvalue;
-
-        /* Find the 'real' password */
-        tmp = avl_find(check_item, DA_USER_PASSWORD);
-        if (tmp)
-                real_password = estrdup(tmp->avp_strvalue);
-        else if (tmp = avl_find(check_item, DA_PASSWORD_LOCATION)) {
-                switch (tmp->avp_lvalue) {
-                case DV_PASSWORD_LOCATION_SQL:
-#ifdef USE_SQL
-                        real_password = rad_sql_pass(radreq, authdata);
-                        if (!real_password)
-                                return auth_nouser;
-#else
-                        radlog_req(L_ERR, radreq,
-                                   _("SQL authentication not available"));
-                        return auth_nouser;
-#endif
-                        break;
-                /*NOTE: add any new location types here */
-                default:
-                        radlog(L_ERR,
-                               _("unknown Password-Location value: %ld"),
-                               tmp->avp_lvalue);
-                        return auth_fail;
-                }
-        }
-
-        /* Process any prefixes/suffixes. */
-        strip_username(1, namepair->avp_strvalue, check_item, name);
-
-        debug(1,
-                ("auth_type=%d, userpass=%s, name=%s, password=%s",
-                 auth_type, userpass, name,
-                 real_password ? real_password : "NONE"));
-
-        switch (auth_type) {
-        case DV_AUTH_TYPE_SYSTEM:
-                debug(1, ("  auth: System"));
-                if (unix_pass(name, userpass) != 0)
-                        result = auth_fail;
-		else
-			result = unix_expiration(name, exp);
-                break;
-		
-        case DV_AUTH_TYPE_PAM:
-#ifdef USE_PAM
-                debug(1, ("  auth: Pam"));
-                /* Provide defaults for authdata */
-                if (authdata == NULL &&
-                    (tmp = avl_find(check_item, DA_PAM_AUTH)) != NULL) {
-                        authdata = tmp->avp_strvalue;
-                }
-                authdata = authdata ? authdata : PAM_DEFAULT_TYPE;
-                if (pam_pass(name, userpass, authdata, user_msg) != 0)
-                        result = auth_fail;
-#else
-                radlog_req(L_ERR, radreq,
-                           _("PAM authentication not available"));
-                result = auth_nouser;
-#endif
-                break;
-
-        case DV_AUTH_TYPE_CRYPT_LOCAL:
-                debug(1, ("  auth: Crypt"));
-                if (real_password == NULL) {
-                        result = auth_fail;
-                        break;
-                }
-                pwlen = strlen(real_password)+1;
-                pwbuf = emalloc(pwlen);
-                if (!md5crypt(userpass, real_password, pwbuf, pwlen))
-                        result = auth_fail;
-                else if (strcmp(real_password, pwbuf) != 0)
-                        result = auth_fail;
-                debug(1,("pwbuf: %s", pwbuf));
-                efree(pwbuf);
-                break;
-                
-        case DV_AUTH_TYPE_LOCAL:
-                debug(1, ("  auth: Local"));
-                /* Local password is just plain text. */
-                if (auth_item->attribute != DA_CHAP_PASSWORD) {
-                        if (real_password == NULL ||
-                            strcmp(real_password, userpass) != 0)
-                                result = auth_fail;
-                        break;
-                }
-
-                /* CHAP: RFC 2865, page 7
-		   The RADIUS server looks up a password based on the
-		   User-Name, encrypts the challenge using MD5 on the
-		   CHAP ID octet, that password, and the CHAP challenge 
-		   (from the CHAP-Challenge attribute if present,
-		   otherwise from the Request Authenticator), and compares
-		   that result to the CHAP-Password.  If they match, the
-		   server sends back an Access-Accept, otherwise it sends
-		   back an Access-Reject. */
-
-		/* Provide some userpass in case authentication fails */
-                strcpy(userpass, "{chap-password}");
-		
-                if (real_password == NULL) {
-                        result = auth_fail;
-                        break;
-                }
-
-		/* Compute the length of the password buffer and
-		   allocate it */
-                length = strlen(real_password);
-		
-                if (tmp = avl_find(radreq->request, DA_CHAP_CHALLENGE)) {
-                        challenge = tmp->avp_strvalue;
-                        challenge_len = tmp->avp_strlength;
-                } else {
-                        challenge = radreq->vector;
-                        challenge_len = AUTH_VECTOR_LEN;
-                }
-
-                pwlen = 1 + length + challenge_len;
-		pwbuf = emalloc(pwlen);
-		
-                ptr = pwbuf;
-                *ptr++ = *auth_item->avp_strvalue;
-                memcpy(ptr, real_password, length);
-                ptr += length;
-                memcpy(ptr, challenge, challenge_len);
-
-		/* Compute the MD5 hash */
-                md5_calc(pw_digest, (u_char*) pwbuf, pwlen);
-                efree(pwbuf);
-		
-                /* Compare them */
-                if (memcmp(pw_digest, auth_item->avp_strvalue + 1,
-			   CHAP_VALUE_LENGTH) != 0)
-                        result = auth_fail;
-                else
-                        strcpy(userpass, real_password);
-                break;
-		
-        default:
-                result = auth_fail;
-                break;
-        }
-
-        if (real_password) {
-                /* just in case: */
-                memset(real_password, 0, strlen(real_password)); 
-                efree(real_password);
-        }
-        return result;
-}
-
 int
-rad_auth_check_username(RADIUS_REQ *radreq, int activefd)
+rad_auth_check_username(radiusd_request_t *radreq, int activefd)
 {
-        VALUE_PAIR *namepair = avl_find(radreq->request, DA_USER_NAME);
+        grad_avp_t *namepair = grad_avl_find(radreq->request->avlist,
+					     DA_USER_NAME);
 
 	log_open(L_AUTH);
 
-        if (avp_null_string_p(namepair)) 
-                radlog_req(L_ERR, radreq, _("No username"));
+        if (grad_avp_null_string_p(namepair)) 
+                grad_log_req(L_ERR, radreq->request, _("No username"));
 	else if (check_user_name(namepair->avp_strvalue)) 
-                radlog_req(L_ERR, radreq, _("Malformed username"));
+                grad_log_req(L_ERR, radreq->request, _("Malformed username"));
 	else
 		return 0;
 
 	/* Process a malformed request */
 	if (auth_reject_malformed_names) 
 		radius_send_reply(RT_ACCESS_REJECT, radreq,
-				  radreq->request,
+				  NULL,
 				  message_text[MSG_ACCESS_DENIED],
 				  activefd);
 	else
-		stat_inc(auth, radreq->ipaddr, num_bad_req);
+		stat_inc(auth, radreq->request->ipaddr, num_bad_req);
 	return -1;
 }
 
 /* Initial step of authentication. */
 int
-rad_auth_init(RADIUS_REQ *radreq, int activefd)
+rad_auth_init(radiusd_request_t *radreq, int activefd)
 {
-	LOCUS loc;
+	grad_locus_t loc;
 	
 	log_open(L_AUTH);
                 
         if (auth_detail)
-                write_detail(radreq, REQ_AUTH_ZERO, "detail.auth");
+                write_detail(radreq, REQ_AUTH_ZERO, R_AUTH);
 
         /*
          * See if the user has access to this huntgroup.
          */
         if (!huntgroup_access(radreq, &loc)) {
-                radlog_req(L_NOTICE, radreq,
-			   _("Access denied by huntgroup %s:%d"),
-			   loc.file, loc.line);
+                grad_log_req(L_NOTICE, radreq->request,
+			     _("Access denied by huntgroup %s:%d"),
+			     loc.file, loc.line);
                 radius_send_reply(RT_ACCESS_REJECT, radreq,
-                                  radreq->request, NULL, activefd);
+                                  radreq->request->avlist, NULL, activefd);
                 return -1;
         }
 
@@ -491,14 +357,14 @@ enum list_id {
 };
 
 typedef struct auth_mach {
-        RADIUS_REQ *req;
-        VALUE_PAIR *user_check;
-        VALUE_PAIR *user_reply;
+        radiusd_request_t *req;
+        grad_avp_t *user_check;
+        grad_avp_t *user_reply;
         int        activefd;
         
-        VALUE_PAIR *namepair;
-        VALUE_PAIR *check_pair;
-        char       userpass[AUTH_STRING_LEN+1];
+        grad_avp_t *namepair;
+        grad_avp_t *check_pair;
+        char       userpass[GRAD_STRING_LENGTH+1];
 
         char       *user_msg;
         struct obstack msg_stack;
@@ -591,21 +457,21 @@ auth_log(AUTH_MACH *m, const char *diag, const char *pass,
 	 const char *reason, const char *addstr)
 {
         if (reason)
-                radlog_req(L_NOTICE, m->req,
-			   "%s [%s%s%s]: %s%s",
-			   diag,
-			   m->namepair->avp_strvalue,
-			   pass ? "/" : "",
-			   pass ? pass : "",
-			   reason,
-			   addstr ? addstr : "");
+                grad_log_req(L_NOTICE, m->req->request,
+			     "%s [%s%s%s]: %s%s",
+			     diag,
+			     m->namepair->avp_strvalue,
+			     pass ? "/" : "",
+			     pass ? pass : "",
+			     reason,
+			     addstr ? addstr : "");
         else
-                radlog_req(L_NOTICE, m->req,
-			   "%s [%s%s%s]",
-			   diag,
-			   m->namepair->avp_strvalue,
-			   pass ? "/" : "",
-			   pass ? pass : "");
+                grad_log_req(L_NOTICE, m->req->request,
+			     "%s [%s%s%s]",
+			     diag,
+			     m->namepair->avp_strvalue,
+			     pass ? "/" : "",
+			     pass ? pass : "");
 }
 
 int
@@ -614,15 +480,15 @@ is_log_mode(AUTH_MACH *m, int mask)
         int mode = log_mode;
         int xmask = 0;
 #ifdef DA_LOG_MODE_MASK
-        VALUE_PAIR *p;
+        grad_avp_t *p;
 
-        for (p = avl_find(m->user_check, DA_LOG_MODE_MASK);
+        for (p = grad_avl_find(m->user_check, DA_LOG_MODE_MASK);
              p;
-             p = p->next ? avl_find(p->next, DA_LOG_MODE_MASK) : NULL)
+             p = p->next ? grad_avl_find(p->next, DA_LOG_MODE_MASK) : NULL)
                 xmask |= p->avp_lvalue;
-        for (p = avl_find(m->req->request, DA_LOG_MODE_MASK);
+        for (p = grad_avl_find(m->req->request->avlist, DA_LOG_MODE_MASK);
              p;
-             p = p->next ? avl_find(p->next, DA_LOG_MODE_MASK) : NULL)
+             p = p->next ? grad_avl_find(p->next, DA_LOG_MODE_MASK) : NULL)
                 xmask |= p->avp_lvalue;
 #endif
         return (mode & ~xmask) & mask;
@@ -642,16 +508,227 @@ auth_finish_msg(AUTH_MACH *m)
                 obstack_grow(&m->msg_stack, m->user_msg, strlen(m->user_msg));
         obstack_1grow(&m->msg_stack, 0);
         return radius_xlate(&m->msg_stack, obstack_finish(&m->msg_stack),
-                            m->req, m->user_reply);
+                            m->req->request, m->user_reply);
+}
+
+/* Check password. */
+static enum auth_status 
+rad_check_password(radiusd_request_t *radreq, AUTH_MACH *m, time_t *exp)
+{
+        char *ptr;
+        char *real_password = NULL;
+        char name[GRAD_STRING_LENGTH];
+        grad_avp_t *auth_item;
+        grad_avp_t *tmp;
+        int auth_type = -1;
+        int length;
+        enum auth_status result = auth_ok;
+        char *authdata = NULL;
+        char pw_digest[GRAD_AUTHENTICATOR_LENGTH];
+        int pwlen;
+        char *pwbuf;
+        char *challenge;
+	int challenge_len;
+	
+        m->userpass[0] = 0;
+
+        /* Process immediate authentication types */
+        if ((tmp = grad_avl_find(m->user_check, DA_AUTH_TYPE)) != NULL)
+                auth_type = tmp->avp_lvalue;
+
+	switch (auth_type) {
+	case DV_AUTH_TYPE_ACCEPT:
+                return auth_ok;
+
+	case DV_AUTH_TYPE_REJECT:
+                return auth_reject;
+
+	case DV_AUTH_TYPE_IGNORE:
+		return auth_ignore;
+	}
+
+        /* Find the password sent by the user. If it's not present,
+           authentication fails. */
+        
+        if (auth_item = grad_avl_find(radreq->request->avlist, DA_CHAP_PASSWORD))
+                auth_type = DV_AUTH_TYPE_LOCAL;
+        else
+                auth_item = grad_avl_find(radreq->request->avlist, DA_USER_PASSWORD);
+        
+        /* Decrypt the password. */
+        if (auth_item) {
+                if (auth_item->avp_strlength == 0)
+                        m->userpass[0] = 0;
+                else
+                        req_decrypt_password(m->userpass, radreq->request,
+                                             auth_item);
+        } else /* if (auth_item == NULL) */
+                return auth_fail;
+        
+        /* Set up authentication data */
+        if ((tmp = grad_avl_find(m->user_check, DA_AUTH_DATA)) != NULL) 
+                authdata = tmp->avp_strvalue;
+
+        /* Find the 'real' password */
+        tmp = grad_avl_find(m->user_check, DA_USER_PASSWORD);
+        if (tmp)
+                real_password = grad_estrdup(tmp->avp_strvalue);
+        else if (tmp = grad_avl_find(m->user_check, DA_PASSWORD_LOCATION)) {
+                switch (tmp->avp_lvalue) {
+                case DV_PASSWORD_LOCATION_SQL:
+#ifdef USE_SQL
+                        real_password = radiusd_sql_pass(radreq, authdata);
+                        if (!real_password)
+                                return auth_nouser;
+			grad_avl_add_pair(&m->user_check,
+					  grad_avp_create_string(DA_USER_PASSWORD, real_password));
+                        break;
+#endif
+                /* NOTE: add any new location types here */
+                default:
+                        grad_log(L_ERR,
+                                 _("unknown Password-Location value: %ld"),
+                                 tmp->avp_lvalue);
+                        return auth_fail;
+                }
+        }
+
+        /* Process any prefixes/suffixes. */
+        strip_username(1, m->namepair->avp_strvalue, m->user_check, name);
+
+        debug(1,
+	      ("auth_type=%d, userpass=%s, name=%s, password=%s",
+	       auth_type, m->userpass, name,
+	       real_password ? real_password : "NONE"));
+
+        switch (auth_type) {
+        case DV_AUTH_TYPE_SYSTEM:
+                debug(1, ("  auth: System"));
+                if (unix_pass(name, m->userpass) != 0)
+                        result = auth_fail;
+		else
+			result = unix_expiration(name, exp);
+                break;
+		
+        case DV_AUTH_TYPE_PAM:
+#ifdef USE_PAM
+                debug(1, ("  auth: Pam"));
+                /* Provide defaults for authdata */
+                if (authdata == NULL &&
+                    (tmp = grad_avl_find(m->user_check, DA_PAM_AUTH)) != NULL) {
+                        authdata = tmp->avp_strvalue;
+                }
+                authdata = authdata ? authdata : PAM_DEFAULT_TYPE;
+                if (pam_pass(name, m->userpass, authdata, &m->user_msg) != 0)
+                        result = auth_fail;
+#else
+                grad_log_req(L_ERR, radreq->request,
+                             _("PAM authentication not available"));
+                result = auth_nouser;
+#endif
+                break;
+
+        case DV_AUTH_TYPE_CRYPT_LOCAL:
+                debug(1, ("  auth: Crypt"));
+                if (real_password == NULL) {
+                        result = auth_fail;
+                        break;
+                }
+                pwlen = strlen(real_password)+1;
+                pwbuf = grad_emalloc(pwlen);
+                if (!grad_md5crypt(m->userpass, real_password, pwbuf, pwlen))
+                        result = auth_fail;
+                else if (strcmp(real_password, pwbuf) != 0)
+                        result = auth_fail;
+                debug(1,("pwbuf: %s", pwbuf));
+                grad_free(pwbuf);
+                break;
+                
+        case DV_AUTH_TYPE_LOCAL:
+                debug(1, ("  auth: Local"));
+                /* Local password is just plain text. */
+                if (auth_item->attribute != DA_CHAP_PASSWORD) {
+                        if (real_password == NULL ||
+                            strcmp(real_password, m->userpass) != 0)
+                                result = auth_fail;
+                        break;
+                }
+
+                /* CHAP: RFC 2865, page 7
+		   The RADIUS server looks up a password based on the
+		   User-Name, encrypts the challenge using MD5 on the
+		   CHAP ID octet, that password, and the CHAP challenge 
+		   (from the CHAP-Challenge attribute if present,
+		   otherwise from the Request Authenticator), and compares
+		   that result to the CHAP-Password.  If they match, the
+		   server sends back an Access-Accept, otherwise it sends
+		   back an Access-Reject. */
+
+		/* Provide some userpass in case authentication fails */
+                strcpy(m->userpass, "{chap-password}");
+		
+                if (real_password == NULL) {
+                        result = auth_fail;
+                        break;
+                }
+
+		/* Compute the length of the password buffer and
+		   allocate it */
+                length = strlen(real_password);
+		
+                if (tmp = grad_avl_find(radreq->request->avlist,
+					DA_CHAP_CHALLENGE)) {
+                        challenge = tmp->avp_strvalue;
+                        challenge_len = tmp->avp_strlength;
+                } else {
+                        challenge = radreq->request->authenticator;
+                        challenge_len = GRAD_AUTHENTICATOR_LENGTH;
+                }
+
+                pwlen = 1 + length + challenge_len;
+		pwbuf = grad_emalloc(pwlen);
+		
+                ptr = pwbuf;
+                *ptr++ = *auth_item->avp_strvalue;
+                memcpy(ptr, real_password, length);
+                ptr += length;
+                memcpy(ptr, challenge, challenge_len);
+
+		/* Compute the MD5 hash */
+                grad_md5_calc(pw_digest, (u_char*) pwbuf, pwlen);
+                grad_free(pwbuf);
+		
+                /* Compare them */
+                if (memcmp(pw_digest, auth_item->avp_strvalue + 1,
+			   GRAD_CHAP_VALUE_LENGTH) != 0)
+                        result = auth_fail;
+                else
+                        strcpy(m->userpass, real_password);
+                break;
+		
+        default:
+		/* Try loadable modules. */
+		result = scheme_try_auth(auth_type, radreq,
+					 m->user_check,
+					 &m->user_reply) ?  auth_fail : auth_ok;
+                break;
+        }
+
+        if (real_password) {
+                /* just in case: */
+                memset(real_password, 0, strlen(real_password)); 
+                grad_free(real_password);
+        }
+        return result;
 }
 
 /* Check if account has expired, and if user may login now. */
 enum auth_status 
 radius_check_expiration(AUTH_MACH *m, time_t *exp)
 {
-        VALUE_PAIR *pair;
+        grad_avp_t *pair;
         
-        if (pair = avl_find(m->user_check, DA_EXPIRATION)) {
+        if (pair = grad_avl_find(m->user_check, DA_EXPIRATION)) {
 		struct timeval tv;
 		gettimeofday(&tv, NULL);
 		if (tv.tv_sec > pair->avp_lvalue)
@@ -664,7 +741,7 @@ radius_check_expiration(AUTH_MACH *m, time_t *exp)
 
 
 int
-rad_authenticate(RADIUS_REQ *radreq, int activefd)
+rad_authenticate(radiusd_request_t *radreq, int activefd)
 {
         enum auth_state oldstate;
         struct auth_state_s *sp;
@@ -679,7 +756,7 @@ rad_authenticate(RADIUS_REQ *radreq, int activefd)
         m.user_msg   = NULL;
         obstack_init(&m.msg_stack);
 
-        m.namepair = avl_find(m.req->request, DA_USER_NAME);
+        m.namepair = grad_avl_find(m.req->request->avlist, DA_USER_NAME);
 
         debug(1, ("auth: %s", m.namepair->avp_strvalue)); 
         m.state = as_init;
@@ -688,11 +765,11 @@ rad_authenticate(RADIUS_REQ *radreq, int activefd)
                 sp = &states[m.state];
                 oldstate = m.state;
                 if (sp->attr) {
-                        VALUE_PAIR *p;
+                        grad_avp_t *p;
                         
                         switch (sp->list) {
                         case L_req:
-                                p = m.req->request;
+                                p = m.req->request->avlist;
                                 break;
                         case L_check:
                                 p = m.user_check;
@@ -703,7 +780,7 @@ rad_authenticate(RADIUS_REQ *radreq, int activefd)
                         default:
                                 abort();
                         }
-                        if (p = avl_find(p, sp->attr))
+                        if (p = grad_avl_find(p, sp->attr))
                                 m.check_pair = p;
                         else {
                                 m.state = sp->next;
@@ -717,12 +794,12 @@ rad_authenticate(RADIUS_REQ *radreq, int activefd)
         }
 
         /* Cleanup */
-        avl_free(m.user_check);
-        avl_free(m.user_reply);
+        grad_avl_free(m.user_check);
+        grad_avl_free(m.user_reply);
         if (m.user_msg)
                 free(m.user_msg);
         obstack_free(&m.msg_stack, NULL);
-        bzero(m.userpass, sizeof(m.userpass));
+        memset(m.userpass, 0, sizeof(m.userpass));
         return 0;
 }
 
@@ -739,18 +816,18 @@ rad_authenticate(RADIUS_REQ *radreq, int activefd)
 void
 sfn_init(AUTH_MACH *m)
 {
-        RADIUS_REQ *radreq = m->req;
-        VALUE_PAIR *pair_ptr;
+        radiusd_request_t *radreq = m->req;
+        grad_avp_t *pair_ptr;
 
 	switch (radreq->server_code) {
 	case RT_ACCESS_REJECT:
-		m->user_check = avp_create_integer(DA_AUTH_TYPE, 
-					           DV_AUTH_TYPE_REJECT);
+		m->user_check = grad_avp_create_integer(DA_AUTH_TYPE, 
+							DV_AUTH_TYPE_REJECT);
 		break;
 
 	case RT_ACCESS_ACCEPT:
-		m->user_check = avp_create_integer(DA_AUTH_TYPE, 
-					           DV_AUTH_TYPE_ACCEPT);
+		m->user_check = grad_avp_create_integer(DA_AUTH_TYPE, 
+							DV_AUTH_TYPE_ACCEPT);
 		break;
 
 	case 0:
@@ -771,7 +848,7 @@ sfn_init(AUTH_MACH *m)
 	 * If the request is processing a menu, service it here.
 	 */
 	if (radreq->server_code == 0
-	    && (pair_ptr = avl_find(m->req->request, DA_STATE)) != NULL
+	    && (pair_ptr = grad_avl_find(m->req->request->avlist, DA_STATE)) != NULL
 	    && strncmp(pair_ptr->avp_strvalue, "MENU=", 5) == 0) {
 		menu_reply(m->req, m->activefd);
 		newstate(as_stop);
@@ -812,20 +889,18 @@ void
 sfn_scheme(AUTH_MACH *m)
 {
 #ifdef USE_SERVER_GUILE
-        VALUE_PAIR *p;
-        VALUE_PAIR *reply;
+        grad_avp_t *p;
+        grad_avp_t *reply = NULL;
         
         if (!use_guile) {
-                radlog_req(L_ERR, m->req,
-                       _("Guile authentication disabled in config"));
+                grad_log_req(L_ERR, m->req->request,
+                             _("Guile authentication disabled in config"));
                 newstate(as_reject_cleanup);
                 return;
         }
 
-        reply = avl_dup(m->user_reply);
-        for (p = avl_find(reply, DA_SCHEME_PROCEDURE);
-             p;
-             p = avl_find(p->next, DA_SCHEME_PROCEDURE)) {
+	grad_avl_move_attr(&reply, &m->user_reply, DA_SCHEME_PROCEDURE);
+        for (p = reply; p; p = p->next) {
                 if (scheme_auth(p->avp_strvalue,
                                 m->req, m->user_check, &m->user_reply)) {
 			auth_log(m,
@@ -837,32 +912,59 @@ sfn_scheme(AUTH_MACH *m)
                         break;
                 }
         }
-        avl_delete(&m->user_reply, DA_SCHEME_PROCEDURE);
-        avl_free(reply);
+        grad_avl_free(reply);
 #else
-        radlog_req(L_ERR, m->req,
-               _("Guile authentication not available"));
+        grad_log_req(L_ERR, m->req,
+                     _("Guile authentication not available"));
         newstate(as_reject_cleanup);
         return;
 #endif
 }
 
+/* Execute an Authentication Failure Trigger, if the one is specified */
+static void
+auth_failure(AUTH_MACH *m)
+{
+	grad_avp_t *pair;
+	char *cmd;
+	struct obstack stk;
+	
+	pair = grad_avl_find(m->user_reply, DA_AUTH_FAILURE_TRIGGER);
+	if (!pair) 
+		return;
+
+	obstack_init(&stk);
+	cmd = util_xlate(&m->msg_stack, pair->avp_strvalue, m->req->request);
+	switch (cmd[0]) {
+	case '(':
+		scheme_eval_boolean_expr(cmd);
+		break;
+
+	case '/':
+		radius_exec_command(cmd);
+		break;
+
+	}
+	obstack_free(&stk, NULL);
+}
+
 void
 sfn_validate(AUTH_MACH *m)
 {
-        RADIUS_REQ *radreq = m->req;
+        radiusd_request_t *radreq = m->req;
 	enum auth_status rc;
 	time_t exp;
-	VALUE_PAIR *pair;
+	grad_avp_t *pair;
 	
-	rc = rad_check_password(radreq,
-				m->user_check, m->namepair,
-				&m->user_msg,
-				m->userpass,
-				&exp);
+	rc = rad_check_password(radreq, m, &exp);
 
-	if (rc == auth_ok) 
+	if (rc == auth_ok) {
+		radiusd_sql_auth_result_query(m->req, 0);
 		rc = radius_check_expiration(m, &exp);
+	} else if (rc == auth_fail) {
+		radiusd_sql_auth_result_query(m->req, 1);
+		auth_failure(m);
+	}
 	
 	switch (rc) {
 	case auth_ok:
@@ -871,9 +973,9 @@ sfn_validate(AUTH_MACH *m)
 	case auth_valid:
 		exp /= SECONDS_PER_DAY;
 		if (warning_seconds != 0 && exp < warning_seconds) {
-			pair = avp_create_integer(DA_PASSWORD_EXPIRE_DAYS,
-						  exp);
-			avl_add_pair(&m->user_reply, pair);
+			pair = grad_avp_create_integer(DA_PASSWORD_EXPIRE_DAYS,
+						       exp);
+			grad_avl_add_pair(&m->user_reply, pair);
 			auth_format_msg(m, MSG_PASSWORD_EXPIRE_WARNING);
 		}
 		break;
@@ -935,7 +1037,7 @@ sfn_validate(AUTH_MACH *m)
 		break;
 
 	default:
-		insist_fail("sfn_validate");
+		grad_insist_fail("sfn_validate");
 	}
 	return;
 }
@@ -956,7 +1058,7 @@ sfn_realmuse(AUTH_MACH *m)
         if (!m->req->realm)
                 return;
         
-        if (rad_check_realm(m->req->realm) == 0)
+        if (radius_mlc_realm(m->req) == 0)
                 return;
         auth_format_msg(m, MSG_REALM_QUOTA);
         auth_log(m, _("Login failed"), NULL,
@@ -967,16 +1069,16 @@ sfn_realmuse(AUTH_MACH *m)
 void
 sfn_simuse(AUTH_MACH *m)
 {
-        char  name[AUTH_STRING_LEN];
+        char  name[GRAD_STRING_LENGTH];
         int rc;
         int count;
 	
         strip_username(strip_names,
                        m->namepair->avp_strvalue, m->user_check, name);
-        rc = rad_check_multi(name, m->req->request,
+        rc = radius_mlc_user(name, m->req,
                              m->check_pair->avp_lvalue, &count);
-        avl_add_pair(&m->user_reply,
-                     avp_create_integer(DA_SIMULTANEOUS_USE, count));
+        grad_avl_add_pair(&m->user_reply,
+			  grad_avp_create_integer(DA_SIMULTANEOUS_USE, count));
         if (!rc)
                 return;
 
@@ -984,22 +1086,22 @@ sfn_simuse(AUTH_MACH *m)
                         (m->check_pair->avp_lvalue > 1) ?
                         MSG_MULTIPLE_LOGIN : MSG_SECOND_LOGIN);
 
-        radlog_req(L_WARN, m->req,
-		   _("Multiple logins: [%s] max. %ld%s"),
-		   m->namepair->avp_strvalue,
-		   m->check_pair->avp_lvalue,
-		   rc == 2 ? _(" [MPP attempt]") : "");
+        grad_log_req(L_WARN, m->req->request,
+		     _("Multiple logins: [%s] max. %ld%s"),
+		     m->namepair->avp_strvalue,
+		     m->check_pair->avp_lvalue,
+		     rc == 2 ? _(" [MPP attempt]") : "");
         newstate(as_reject_cleanup);
 }
 
-static UINT4
-set_session_timeout(AUTH_MACH *m, UINT4 val)
+static grad_uint32_t
+set_session_timeout(AUTH_MACH *m, grad_uint32_t val)
 {
-	VALUE_PAIR *p;
+	grad_avp_t *p;
 	
-        if (!(p = avl_find(m->user_reply, DA_SESSION_TIMEOUT))) {
-                p = avp_create_integer(DA_SESSION_TIMEOUT, val);
-                avl_add_pair(&m->user_reply, p);
+        if (!(p = grad_avl_find(m->user_reply, DA_SESSION_TIMEOUT))) {
+                p = grad_avp_create_integer(DA_SESSION_TIMEOUT, val);
+                grad_avl_add_pair(&m->user_reply, p);
         } else if (p->avp_lvalue > val)
 		p->avp_lvalue = val;
         return p->avp_lvalue;
@@ -1019,16 +1121,16 @@ sfn_time(AUTH_MACH *m)
                  * User called outside allowed time interval.
                  */
                 auth_format_msg(m, MSG_TIMESPAN_VIOLATION);
-                radlog_req(L_ERR,
-			   m->req,
-			   _("Outside allowed timespan (%s)"),
-			   m->check_pair->avp_strvalue);
+                grad_log_req(L_ERR,
+			     m->req->request,
+			     _("Outside allowed timespan (%s)"),
+			     m->check_pair->avp_strvalue);
                 newstate(as_reject_cleanup);
         } else if (rc == 0) {
                 /*
                  * User is allowed, but set Session-Timeout.
                  */
-                UINT4 to = set_session_timeout(m, rest);
+                grad_uint32_t to = set_session_timeout(m, rest);
                 debug(2, ("user %s, span %s, timeout %d, real timeout %d",
                           m->namepair->avp_strvalue,
                           m->check_pair->avp_strvalue,
@@ -1040,13 +1142,14 @@ sfn_time(AUTH_MACH *m)
 void
 sfn_ipaddr(AUTH_MACH *m)
 {
-        VALUE_PAIR *p, *tmp, *pp;
+        grad_avp_t *p, *tmp, *pp;
         
         /* Assign an IP if necessary */
-        if (!avl_find(m->user_reply, DA_FRAMED_IP_ADDRESS)) {
-                if (p = avl_find(m->req->request, DA_FRAMED_IP_ADDRESS)) {
+        if (!grad_avl_find(m->user_reply, DA_FRAMED_IP_ADDRESS)) {
+                if (p = grad_avl_find(m->req->request->avlist,
+				      DA_FRAMED_IP_ADDRESS)) {
                         /* termserver hint */
-                        avl_add_pair(&m->user_reply, avp_dup(p));
+                        grad_avl_add_pair(&m->user_reply, grad_avp_dup(p));
                 }
         }
         
@@ -1056,13 +1159,13 @@ void
 sfn_exec_wait(AUTH_MACH *m)
 {
 	int rc;
-	VALUE_PAIR *p;
-	VALUE_PAIR *repl = NULL;
+	grad_avp_t *p;
+	grad_avp_t *repl = NULL;
 	
 	for (p = m->check_pair;
              p;
-             p = avl_find(p->next, DA_EXEC_PROGRAM_WAIT)) {
-		VALUE_PAIR *tail = NULL;
+             p = grad_avl_find(p->next, DA_EXEC_PROGRAM_WAIT)) {
+		grad_avp_t *tail = NULL;
 
 		if (m->user_reply)
 			for (tail = m->user_reply; tail->next;
@@ -1103,9 +1206,9 @@ sfn_exec_wait(AUTH_MACH *m)
 
 			p = tail->next;
 			tail->next = NULL;
-			avl_free(m->user_reply);
+			grad_avl_free(m->user_reply);
 			m->user_reply = p;
-			if (!avl_find(m->user_reply, DA_REPLY_MESSAGE))
+			if (!grad_avl_find(m->user_reply, DA_REPLY_MESSAGE))
 				auth_format_msg(m, MSG_ACCESS_DENIED);
 			
 			break;
@@ -1116,11 +1219,11 @@ sfn_exec_wait(AUTH_MACH *m)
 void
 sfn_exec_nowait(AUTH_MACH *m)
 {
-	VALUE_PAIR *p;
+	grad_avp_t *p;
 	
 	for (p = m->check_pair;
              p;
-             p = avl_find(p->next, DA_EXEC_PROGRAM)) {
+             p = grad_avl_find(p->next, DA_EXEC_PROGRAM)) {
 		
 		radius_exec_program(m->check_pair->avp_strvalue,
 				    m->req, NULL, 0);
@@ -1146,7 +1249,7 @@ sfn_cleanup_cbkid(AUTH_MACH *m)
         int *ip;
 
         for (ip = delete_pairs; *ip; ip++)
-                avl_delete(&m->user_reply, *ip);
+                grad_avl_delete(&m->user_reply, *ip);
 }
 
 void
@@ -1160,7 +1263,7 @@ sfn_menu_challenge(AUTH_MACH *m)
         snprintf(state_value, sizeof(state_value),
                    "MENU=%s", m->check_pair->avp_strvalue);
         radius_send_challenge(m->req, msg, state_value, m->activefd);
-        efree(msg);
+        grad_free(msg);
 	
         debug(1,
               ("sending challenge (menu %s) to %s",
@@ -1192,7 +1295,7 @@ sfn_ack(AUTH_MACH *m)
 void
 sfn_reject_cleanup(AUTH_MACH *m)
 {
-	avl_free(m->user_reply);
+	grad_avl_free(m->user_reply);
 	m->user_reply = NULL;
 	newstate(as_reject);
 }
@@ -1210,32 +1313,35 @@ sfn_reject(AUTH_MACH *m)
 }
 
 void
-req_decrypt_password(char *password, RADIUS_REQ *req, VALUE_PAIR *pair)
+req_decrypt_password(char *password, grad_request_t *req, grad_avp_t *pair)
 {
-        NAS *nas;
+        grad_nas_t *nas;
         char *s;
         
         if (!pair) {
-                pair = avl_find(req->request, DA_USER_PASSWORD);
+                pair = grad_avl_find(req->avlist, DA_USER_PASSWORD);
                 if (!pair)
                         return;
         }
 
-	if (pair->prop & AP_ENCRYPT_RFC2138) {
+	if (pair->prop & GRAD_AP_ENCRYPT_RFC2138) {
 		/* Determine whether we need to use broken decoding */
-		nas = nas_request_to_nas(req);
+		nas = grad_nas_request_to_nas(req);
 		if (nas
-		    && (s = envar_lookup(nas->args, "broken_pass")) != NULL
+		    && (s = grad_envar_lookup(nas->args, "broken_pass")) != NULL
 		    && s[0] == '1')
-			decrypt_password_broken(password, pair,
-						req->vector, req->secret);
+			grad_decrypt_password_broken(password, pair,
+						     req->authenticator,
+						     req->secret);
 		else
-			decrypt_password(password, pair,
-					 req->vector, req->secret);
-	} else if (pair->prop & AP_ENCRYPT_RFC2868) {
+			grad_decrypt_password(password, pair,
+					      req->authenticator,
+					      req->secret);
+	} else if (pair->prop & GRAD_AP_ENCRYPT_RFC2868) {
 		u_char tag; /* FIXME: not accessible for user */
-		decrypt_tunnel_password(password, 
-					&tag, pair,
-					req->vector, req->secret);
+		grad_decrypt_tunnel_password(password, 
+					     &tag, pair,
+					     req->authenticator,
+					     req->secret);
 	}
 }
