@@ -42,6 +42,8 @@
 const char *argp_program_version = "radiusd (" PACKAGE ") " VERSION;
 static char doc[] = N_("GNU radius daemon");
 
+#define SHOW_DEFAULTS_OPTION 256
+
 static struct argp_option options[] = {
         {NULL, 0, NULL, 0,
          N_("radiusd specific switches:"), 0},
@@ -77,6 +79,8 @@ static struct argp_option options[] = {
          N_("Log authentications"), 0},
         {"log-auth-pass", 'z', NULL, 0,
          N_("Log users' passwords"), 0},
+	{"show-defaults", SHOW_DEFAULTS_OPTION, NULL, 0,
+	 N_("Show compilation defaults"), 0},
         {NULL, 0, NULL, 0, NULL, 0}
 };
 
@@ -247,7 +251,7 @@ parse_opt(int key, char *arg, struct argp_state *state)
 		
         case 'x':
                 x_debug_spec = optarg;
-                set_debug_levels(optarg);
+                grad_set_debug_levels(optarg);
                 break;
 		
         case 'y':
@@ -257,6 +261,10 @@ parse_opt(int key, char *arg, struct argp_state *state)
         case 'z':
                 log_mode |= RLOG_AUTH_PASS;
                 break;
+
+	case SHOW_DEFAULTS_OPTION:
+		show_compilation_defaults();
+		exit(0);
 		
         default:
                 return ARGP_ERR_UNKNOWN;
@@ -380,8 +388,8 @@ radiusd_postconfig_hook(void *a ARG_UNUSED, void *b ARG_UNUSED)
 
 static void
 daemon_postconfig_hook(void *a ARG_UNUSED, void *b ARG_UNUSED)
-{        system_acct_init();
- 
+{
+	system_acct_init();
 }
 
 void
@@ -401,6 +409,19 @@ radiusd_setup()
 	snmp_port = get_port_number("snmp", "udp", 161);
 #endif
         srand(time(NULL));
+	
+#ifdef HAVE_CRYPT_SET_FORMAT
+	/* MD5 hashes are handled by libgnuradius function md5crypt(). To
+	   handle DES hashes it falls back to system crypt(). The behaviour
+	   of the latter on FreeBSD depends upon a 'default format'. so e.g.
+	   crypt() may generate MD5 hashes even if presented with a valid
+	   MD5 salt.
+	   
+	   To make sure this does not happen, we need to set the default
+	   crypt() format. */
+	
+	crypt_set_format("des");
+#endif
 
 	/* Register radiusd hooks first. This ensures they will be
 	   executed after all other hooks */
@@ -408,6 +429,7 @@ radiusd_setup()
 	radiusd_set_postconfig_hook(radiusd_postconfig_hook, NULL, 0);
 
 	rewrite_init();
+	dynload_init();
         snmp_init(0, 0, (snmp_alloc_t)grad_emalloc, (snmp_free_t)grad_free);
 	mlc_init();
 	sql_init();
@@ -713,26 +735,25 @@ radiusd_close_channel(int fd)
 }
 
 void
-radiusd_cleanup()
+radiusd_collect_children()
 {
 	pid_t pid;
 	int status;
-	char buffer[128];
-	
-        for (;;) {
 
+	for (;;) {
 		pid = waitpid((pid_t)-1, &status, WNOHANG);
-                if (pid <= 0)
-                        break;
-
-		format_exit_status(buffer, sizeof buffer, status);
-		grad_log(L_NOTICE, _("child %lu %s"),
-		         (unsigned long) pid, buffer);
-
-		rpp_remove(pid);
+		if (pid <= 0)
+			break;
+		rpp_status_changed(pid, status);
 	}
 }
 
+void
+radiusd_cleanup()
+{
+	rpp_collect_exited ();
+}
+
 void
 radiusd_restart()
 {
@@ -950,6 +971,7 @@ sig_handler(int sig)
 		break;
 
 	case SIGCHLD:
+		radiusd_collect_children ();
 		daemon_command = CMD_CLEANUP;
 		break;
 
@@ -1091,7 +1113,7 @@ udp_open(int type, grad_uint32_t ipaddr, int port, int nonblock)
         s.sin_addr.s_addr = htonl(ipaddr);
         s.sin_port = htons(port);
 	if (p = input_find_channel(radius_input, "udp", &s)) {
-		char buffer[DOTTED_QUAD_LEN];
+		char buffer[GRAD_IPV4_STRING_LENGTH];
 		grad_log(L_ERR,
 		         _("socket %s:%d is already assigned for %s"),
 		         grad_ip_iptostr(ipaddr, buffer),
@@ -1376,7 +1398,7 @@ struct cfg_stmt auth_stmt[] = {
 struct cfg_stmt acct_stmt[] = {
 	{ "port", CS_STMT, NULL, cfg_get_port, &acct_port, NULL, NULL },
 	{ "listen", CS_STMT, NULL, rad_cfg_listen_acct, NULL, NULL, NULL },
-	{ "forward", CS_STMT, NULL, rad_cfg_forward_auth, NULL, NULL, NULL },
+	{ "forward", CS_STMT, NULL, rad_cfg_forward_acct, NULL, NULL, NULL },
 	{ "max-requests", CS_STMT, NULL,
 	  cfg_get_integer, &request_class[R_ACCT].max_requests,
 	  NULL, NULL },
@@ -1421,6 +1443,8 @@ struct cfg_stmt config_syntax[] = {
 	{ "rewrite", CS_BLOCK, rewrite_stmt_term, NULL, NULL, rewrite_stmt, NULL },
 	{ "filters", CS_BLOCK, filters_stmt_term, NULL, NULL, filters_stmt,
 	  NULL },
+	{ "loadable-modules", CS_BLOCK, dynload_stmt_term, NULL, NULL,
+	  dynload_stmt, NULL },
 #ifdef USE_DBM
 	{ "usedbm", CS_STMT, NULL, cfg_get_boolean, &use_dbm, NULL, NULL },
 #endif
