@@ -1,5 +1,5 @@
 /* This file is part of GNU Radius
-   Copyright (C) 2000,2002,2003 Free Software Foundation, Inc.
+   Copyright (C) 2000,2002,2003,2004 Free Software Foundation, Inc.
 
    Written by Sergey Poznyakoff
 
@@ -17,22 +17,20 @@
    along with GNU Radius; if not, write to the Free Software Foundation, 
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
-#define RADIUS_MODULE_RADUTIL_C
-
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
 
 #include <stdlib.h>
 #include <ctype.h>
+
 #include <sysdep.h>
 #include <radiusd.h>
-#include <obstack1.h>
 #include <rewrite.h>
 
 /* obstack_grow with quoting of potentially dangerous characters */
 static void
-obstack_grow_quoted(struct obstack *obp, char *str, int len)
+obstack_grow_escaped(struct obstack *obp, char *str, int len)
 {
         for (; len > 0; len--, str++) {
                 switch (*str) {
@@ -86,22 +84,22 @@ obstack_grow_quoted(struct obstack *obp, char *str, int len)
  * others.
  */
 static void
-attr_to_str(struct obstack *obp, RADIUS_REQ *req, VALUE_PAIR *pairlist,
-            DICT_ATTR  *attr, char *defval)
+attr_to_str(struct obstack *obp, grad_request_t *req, grad_avp_t *pairlist,
+            grad_dict_attr_t  *attr, char *defval, int escape)
 {
-        VALUE_PAIR *pair;
+        grad_avp_t *pair;
         int len;
-        char tmp[AUTH_STRING_LEN + 1];
+        char tmp[GRAD_STRING_LENGTH + 1];
         char *str;
 	
         if (!attr) {
-                radlog_req(L_ERR, req, "attribute not found");
+                grad_log_req(L_ERR, req, "attribute not found");
                 return;
         }
         
-        if ((pair = avl_find(pairlist, attr->value)) == NULL) {
+        if ((pair = grad_avl_find(pairlist, attr->value)) == NULL) {
                 if (!defval) {
-                        if (attr->type == TYPE_STRING)
+                        if (attr->type == GRAD_TYPE_STRING)
                                 defval = "-";
                         else
                                 defval = "-0";
@@ -117,28 +115,28 @@ attr_to_str(struct obstack *obp, RADIUS_REQ *req, VALUE_PAIR *pairlist,
                 case '?':
                         if (*defval == 0)
                                 defval = "Attribute is not present";
-                        radlog_req(L_ERR, req, "%s: %s",
-				   attr->name, defval);
+                        grad_log_req(L_ERR, req, "%s: %s",
+				     attr->name, defval);
                         break;
                 case '=':
                         if (pairlist) {
-				LOCUS loc;
+				grad_locus_t loc;
 				loc.file = __FILE__; /*FIXME*/
 				loc.line = __LINE__;
-                                pair = install_pair(&loc,
-                                                    attr->name,
-                                                    OPERATOR_EQUAL,
-                                                    defval);
+                                pair = grad_create_pair(&loc,
+							attr->name,
+							grad_operator_equal,
+							defval);
                                 if (pair)
-                                        avl_add_list(&pairlist, pair);
+                                        grad_avl_add_list(&pairlist, pair);
                         }
                         break;
                 default:
                         if (defval)
-                                radlog(L_ERR, "invalid : substitution: %s",
-                                       defval);
+                                grad_log(L_ERR, "invalid : substitution: %s",
+                                         defval);
                         else
-                                radlog(L_ERR, "null : substitution");
+                                grad_log(L_ERR, "null : substitution");
                         break;
                 }
                 
@@ -151,69 +149,84 @@ attr_to_str(struct obstack *obp, RADIUS_REQ *req, VALUE_PAIR *pairlist,
                 return;
         }
 
-        tmp[AUTH_STRING_LEN-1] = 0;
+        tmp[GRAD_STRING_LENGTH-1] = 0;
         switch (attr->type) {
-        case TYPE_STRING:
-                if ((attr->prop & AP_ENCRYPT) && req) {
+        case GRAD_TYPE_STRING:
+                if ((attr->prop & GRAD_AP_ENCRYPT) && req) {
                         req_decrypt_password(tmp, req, pair);
 			str = tmp;
 		} else
 			str = pair->avp_strvalue;
 
-		if (attr->prop & AP_BINARY_STRING) 
+		if (attr->prop & GRAD_AP_BINARY_STRING) 
 			len = pair->avp_strlength;
 		else
 			/* strvalue might include terminating zero
 			   character, so we need to recalculate it */
 			len = strlen(str);
-		obstack_grow_quoted(obp, str, len);
+		if (escape)
+			obstack_grow_escaped(obp, str, len);
+		else
+			obstack_grow(obp, str, len);
                 break;
 		
-        case TYPE_INTEGER:
-                snprintf(tmp, sizeof(tmp), "%lu", pair->avp_lvalue);
+        case GRAD_TYPE_INTEGER:
+	{
+		grad_dict_value_t *dval;
+		
+                if (escape && (pair->prop & GRAD_AP_TRANSLATE))
+                        dval = grad_value_lookup(pair->avp_lvalue, pair->name);
+                else
+                        dval = NULL;
+                
+                if (!dval) {
+			snprintf(tmp, sizeof(tmp), "%lu", pair->avp_lvalue);
+			len = strlen(tmp);
+			obstack_grow(obp, tmp, len);
+		} else 
+			obstack_grow(obp, dval->name, strlen (dval->name));
+		break;
+	}
+		
+        case GRAD_TYPE_IPADDR:
+                grad_ip_iptostr(pair->avp_lvalue, tmp);
                 len = strlen(tmp);
                 obstack_grow(obp, tmp, len);
                 break;
 		
-        case TYPE_IPADDR:
-                ip_iptostr(pair->avp_lvalue, tmp);
-                len = strlen(tmp);
-                obstack_grow(obp, tmp, len);
-                break;
-		
-        case TYPE_DATE:
+        case GRAD_TYPE_DATE:
                 snprintf(tmp, sizeof(tmp), "%ld", pair->avp_lvalue);
                 len = strlen(tmp);
                 obstack_grow(obp, tmp, len);
                 break;
 		
         default:
-                radlog(L_CRIT,
-                    _("INTERNAL ERROR (%s:%d): attribute %d has bad type (%d)"),
-                    __FILE__, __LINE__,
-                    attr->value, attr->type);
+                grad_log(L_CRIT,
+                         _("INTERNAL ERROR (%s:%d): attribute %d has bad type (%d)"),
+                         __FILE__, __LINE__,
+                         attr->value, attr->type);
                 break;
         }
 }
 
 static void
-curtime_to_str(struct obstack *obp, VALUE_PAIR *request, int gmt)
+curtime_to_str(struct obstack *obp, grad_avp_t *request, int gmt)
 {
         time_t curtime;
         struct tm *tm, tms;
-        VALUE_PAIR *pair;
-        char tbuf[AUTH_STRING_LEN];
+        grad_avp_t *pair;
+        char tbuf[GRAD_STRING_LENGTH];
         int len;
         
         curtime = time(NULL);
-        if (pair = avl_find(request, DA_ACCT_DELAY_TIME))
+        if (pair = grad_avl_find(request, DA_ACCT_DELAY_TIME))
                 curtime -= pair->avp_lvalue;
         if (gmt)
                 tm = gmtime(&curtime);
         else
                 tm = localtime_r(&curtime, &tms);
                                 
-        len = strftime(tbuf, AUTH_STRING_LEN, "%Y-%m-%d %H:%M:%S", tm);
+        len = strftime(tbuf, GRAD_STRING_LENGTH, "%Y-%m-%d %H:%M:%S", tm);
         obstack_grow(obp, tbuf, len);
 }
 
@@ -223,21 +236,21 @@ curtime_to_str(struct obstack *obp, VALUE_PAIR *request, int gmt)
  * attr_to_str)
  */
 static void
-attrno_to_str(struct obstack *obp, RADIUS_REQ *req, VALUE_PAIR *pairlist,
-              int attr_no, char *defval)
+attrno_to_str(struct obstack *obp, grad_request_t *req, grad_avp_t *pairlist,
+              int attr_no, char *defval, int escape_str)
 {
         attr_to_str(obp, req, pairlist,
-		    attr_number_to_dict(attr_no), defval);
+		    grad_attr_number_to_dict(attr_no), defval, escape_str);
 }
 
-static DICT_ATTR *
+static grad_dict_attr_t *
 parse_dict_attr(char *p, char **endp, char **defval)
 {
-        char namebuf[MAX_DICTNAME];
+        char namebuf[GRAD_MAX_DICTNAME];
         
         *defval = NULL;
         if (isdigit(*p)) {
-                return attr_number_to_dict(strtol(p, endp, 10));
+                return grad_attr_number_to_dict(strtol(p, endp, 10));
         }
 
         if (*p == '{') {
@@ -266,26 +279,27 @@ parse_dict_attr(char *p, char **endp, char **defval)
                                 return NULL;
 
                         size = p - start + 1;
-                        *defval = emalloc(size);
+                        *defval = grad_emalloc(size);
                         memcpy(*defval, start, size-1);
                         (*defval)[size] = 0;
                 }
                 *endp = p + 1;
-                return attr_name_to_dict(namebuf);
+                return grad_attr_name_to_dict(namebuf);
         }
         *endp = p;
         return NULL;
 }
 
 void
-radius_xlate0(struct obstack *obp, char *str, RADIUS_REQ *req,
-              VALUE_PAIR *reply)
+radius_xlate0(struct obstack *obp, char *str, grad_request_t *req,
+              grad_avp_t *reply)
 {
         int c;
         char *p;
-        DICT_ATTR *da;
+        grad_dict_attr_t *da;
         char *defval;
-
+	int escape;
+	
         for (p = str; *p; ) {
                 switch (c = *p++) {
                 default:
@@ -305,77 +319,96 @@ radius_xlate0(struct obstack *obp, char *str, RADIUS_REQ *req,
                                 obstack_1grow(obp, c);
                                 break;
                         }
+			escape = (p > str+1 && (p[-2] == '\'' || p[-2] == '"'));
                         switch (c = *p++) {
                         case '%':
                                 obstack_1grow(obp, c);
                                 break;
 				
                         case 'D':
-                                curtime_to_str(obp, req->request, 0);
+                                curtime_to_str(obp, req->avlist, 0);
                                 break;
 				
                         case 'G':
-                                curtime_to_str(obp, req->request, 1);
+                                curtime_to_str(obp, req->avlist, 1);
                                 break;
 				
                         case 'f': /* Framed IP address */
                                 attrno_to_str(obp, NULL, reply,
-                                              DA_FRAMED_IP_ADDRESS, NULL);
+                                              DA_FRAMED_IP_ADDRESS, NULL,
+					      escape);
                                 break;
 				
                         case 'n': /* NAS IP address */
-                                attrno_to_str(obp, req, req->request,
-                                              DA_NAS_IP_ADDRESS, NULL);
+                                attrno_to_str(obp, req,
+					      req->avlist,
+                                              DA_NAS_IP_ADDRESS, NULL,
+					      escape);
                                 break;
 				
                         case 't': /* MTU */
                                 attrno_to_str(obp, NULL, reply,
-                                              DA_FRAMED_MTU, NULL);
+                                              DA_FRAMED_MTU, NULL,
+					      escape);
                                 break;
 				
                         case 'p': /* Port number */
-                                attrno_to_str(obp, req, req->request,
-                                              DA_NAS_PORT_ID, NULL);
+                                attrno_to_str(obp, req, req->avlist,
+                                              DA_NAS_PORT_ID, NULL,
+					      escape);
                                 break;
 				
                         case 'u': /* User name */
-                                attrno_to_str(obp, req, req->request,
-                                              DA_USER_NAME, NULL);
+                                attrno_to_str(obp, req, req->avlist,
+                                              DA_USER_NAME, NULL,
+					      escape);
                                 break;
 				
                         case 'c': /* Callback-Number */
                                 attrno_to_str(obp, NULL, reply,
-                                              DA_CALLBACK_NUMBER, NULL);
+                                              DA_CALLBACK_NUMBER, NULL,
+					      escape);
                                 break;
 				
                         case 'i': /* Calling station ID */
-                                attrno_to_str(obp, req, req->request,
-                                              DA_CALLING_STATION_ID, NULL);
+                                attrno_to_str(obp, req, req->avlist,
+                                              DA_CALLING_STATION_ID, NULL,
+					      escape);
                                 break;
 				
                         case 'a': /* Protocol: SLIP/PPP */
                                 attrno_to_str(obp, NULL, reply,
-                                              DA_FRAMED_PROTOCOL, NULL);
+                                              DA_FRAMED_PROTOCOL, NULL,
+					      escape);
                                 break;
 				
                         case 's': /* Speed */
-                                attrno_to_str(obp, req, req->request,
-                                              DA_CONNECT_INFO, NULL);
+                                attrno_to_str(obp, req, req->avlist,
+                                              DA_CONNECT_INFO, NULL,
+					      escape);
                                 break;
 				
                         case 'C':
-                                /* Check pair */
+				if (*p == '\\') {
+					escape = 1;
+					p++;
+				} 
+                                /* Request pair */
                                 da = parse_dict_attr(p, &p, &defval);
-                                attr_to_str(obp, req, req->request,
-                                            da, defval);
-                                efree(defval);
+                                attr_to_str(obp, req, req->avlist,
+                                            da, defval, escape);
+                                grad_free(defval);
                                 break;
 				
                         case 'R':
+				if (*p == '\\') {
+					escape = 1;
+					p++;
+				}
                                 /* Reply pair */
                                 da = parse_dict_attr(p, &p, &defval);
                                 attr_to_str(obp, NULL,
-                                            reply, da, defval);
+                                            reply, da, defval, escape);
                                 break;
 				
                         default:                                        
@@ -430,17 +463,44 @@ end:
 }
 
 char *
-radius_xlate(struct obstack *obp, char *str, RADIUS_REQ *req, VALUE_PAIR *reply)
+radius_xlate(struct obstack *obp, char *str,
+	     grad_request_t *req, grad_avp_t *reply)
 {
         radius_xlate0(obp, str, req, reply);
         obstack_1grow(obp, 0);
         return obstack_finish(obp);
 }
 
-static void
-pair_set_value(VALUE_PAIR *p, Datatype type, Datum *datum)
+char *
+util_xlate(struct obstack *sp, char *fmt, grad_request_t *radreq)
 {
-	efree(p->avp_strvalue);
+	char *str;
+	
+	if (fmt[0] == '=') {
+		Datatype type;
+		Datum datum;
+
+		/*FIXME: Should be compiled!*/
+		if (rewrite_interpret(fmt+1, radreq, &type, &datum)) 
+			return NULL;
+		if (type != String) {
+			grad_log(L_ERR, "%s: %s",
+			         fmt+1, _("wrong return type"));
+			return NULL;
+		}
+		obstack_grow(sp, datum.sval, strlen(datum.sval)+1);
+		grad_free(datum.sval);
+		str = obstack_finish(sp);
+	} else {
+		str = radius_xlate(sp, fmt, radreq, NULL);
+	}
+	return str;
+}
+
+static void
+pair_set_value(grad_avp_t *p, Datatype type, Datum *datum)
+{
+	grad_free(p->avp_strvalue);
 	switch (type) {
 	case Integer:
 		p->avp_lvalue = datum->ival;
@@ -452,13 +512,13 @@ pair_set_value(VALUE_PAIR *p, Datatype type, Datum *datum)
 		break;
 		
 	default:
-		insist_fail("bad Datatype");
+		grad_insist_fail("bad Datatype");
 	}
-	p->eval_type = eval_const;
+	p->eval_type = grad_eval_const;
 }
 
 int
-radius_eval_avl(RADIUS_REQ *req, VALUE_PAIR *p)
+radius_eval_avl(radiusd_request_t *req, grad_avp_t *p)
 {
 	int errcnt = 0;
 	for (; p; p = p->next) {
@@ -466,21 +526,21 @@ radius_eval_avl(RADIUS_REQ *req, VALUE_PAIR *p)
 		Datum datum;
 		
 		switch (p->eval_type) {
-		case eval_const:
+		case grad_eval_const:
 			break;
 			
-		case eval_interpret:
+		case grad_eval_interpret:
 			if (rewrite_interpret(p->avp_strvalue,
-					      req, &type, &datum)) {
+					      req->request, &type, &datum)) {
                                 errcnt++;
                                 continue;
                         }
 			pair_set_value(p, type, &datum);
 			break;
 
-		case eval_compiled:
-			if (rewrite_eval(p->avp_strvalue,
-					 req, &type, &datum)) {
+		case grad_eval_compiled:
+ 			if (rewrite_eval(p->avp_strvalue,
+					 req->request, &type, &datum)) {
                                 errcnt++;
                                 continue;
                         }
@@ -488,10 +548,30 @@ radius_eval_avl(RADIUS_REQ *req, VALUE_PAIR *p)
 			break;
 
 		default:
-			insist_fail("bad eval_type");
+			grad_insist_fail("bad eval_type");
                 }
         }
 
 	return errcnt++;
 }
 
+radiusd_request_t *
+radiusd_request_alloc(grad_request_t *req)
+{
+	radiusd_request_t *ret = grad_emalloc(sizeof(*ret));
+	ret->request = req;
+	return ret;
+}
+
+void
+radiusd_request_free(radiusd_request_t *radreq)
+{
+	grad_list_destroy(&radreq->locus_list, NULL, NULL);
+	grad_free(radreq->remote_user);
+        grad_avl_free(radreq->reply_pairs);
+        grad_free(radreq->reply_msg);
+        grad_avl_free(radreq->server_reply);
+	grad_request_free(radreq->request);
+	grad_free(radreq);
+}
+
